program Get_critical_point(input, output);

{$APPTYPE CONSOLE}

uses SysUtils, Math;

{ ------------------------------------------------------------------- }
{     Find the critical point for the power calculation.  @          }
{     Max acc t1 for Poisson sequence (step change-point model).      }
{ ------------------------------------------------------------------- }
{     This program can be compiled by Delphi(R).                      }
{     The executable code and the latest version are available at :   }
{        https://corec.meisei-u.ac.jp/labs/hirotsu/                   }
{ ------------------------------------------------------------------- }
{                    Notes and Limitations                            }
{ ------------------------------------------------------------------- }
{     Global variables C1 & C2 are used to avoid stack overflow.      }
{     Maximum length of Poisson sequence is 100.                      }
{     Extended precision is used                                      }
{     & the maximum number of total frequency is 1754.                }
{       (If you use double precision, max number is 170.)             }
{ ------------------------------------------------------------------- }
{                                 written by Hal.Tsuruta  2017/06/02  }
{ ------------------------------------------------------------------- }

{ --- Declarations -------------------------------------------------- }

  const  max_sequence  =                 100;
         max_factorial =                1754;      { max of factorial }
         epsilon       =           0.0000001;
         epsilon2      =              0.0001;
         infinity      =  1.797693134862E308;

  type array_type        = array[0..max_sequence]  of integer;
       array_type_R      = array[0..max_sequence]  of extended;
       array_type_long_R = array[0..max_factorial] of extended;

       Ck_type = record
                   p        : extended;
                   feasible : boolean;
                 end;

  var  fi, fo       : text;
       filename     : string;

       J,
       a,  Ya       : integer;

       alpha,
       alpha1,
       alpha2,
       cp, cp1, cp2,
       p            : extended;

       y            : array_type;
       C1, C2       : array[0..max_factorial] of Ck_type; { global var }
       LFact        : array_type_long_R;

  label last_line;

{ ------------------------------------------------------------------- }
     function Imin2( x1, x2 : integer ) : integer;
{ ------------------------------------------------------------------- }
     begin
       if x1 < x2 then Imin2 := x1
                  else Imin2 := x2;
     end;

{ ------------------------------------------------------------------- }
     function feasible( Y1, Y2 : integer ) : boolean;
{ ------------------------------------------------------------------- }
     var yk : integer;

     begin
       yk := Y2 - Y1;

       if (yk >= 0) then feasible := True
                    else feasible := False;
     end;

{ ------------------------------------------------------------------- }
     function fk( Y1, Y2  : integer ) : extended;
{ ------------------------------------------------------------------- }
{         NB: C1 & C2 are gloabal variables                           }
{ ------------------------------------------------------------------- }

     begin
       if feasible(Y1, Y2) and ((C1[Y1].feasible and C2[Y2].feasible)) then
         fk := exp( Ln(C1[Y1].p) - Ln(C2[Y2].p) - LFact[Y2 - Y1] )
       else
         fk := 0;
     end;

{ ------------------------------------------------------------------- }
     function get_F( a   : integer;
                     cp  : extended;
                     y   : array_type ) : extended;
{ ------------------------------------------------------------------- }
  var k,   J,
      Y1,  Y2,   Ya,
      low_Y1, up_Y1,
      low_Y2, up_Y2,
      max_k0          : integer;

      max_t0,     t,
      lambda_hat, sum : extended;

      Yk,  low,  up   : array_type;
      E0              : array_type_R;
      Fp1, Fp2        : array_type_long_R;

{ ----  Start of get_F  --------------------------------------------- }

  begin
    for J:=0 to max_factorial do begin
      C1[J].p:=0;  C1[J].feasible:=False;  Fp1[J]:=0;
      C2[J].p:=0;  C2[J].feasible:=False;  Fp2[J]:=0;
    end;

{ --- get accumulated sum Yk's of the observed sequence y ----------- }

                     Yk[0] := 0;
    for k:=1 to a do Yk[k] := Yk[k-1] + y[k];
                     Ya    := Yk[a];

{ --- set the range of Yk's ----------------------------------------- }

    for k:=1 to (a-1) do begin
      low[k] :=  0;
      up[k]  := Ya;
    end;

    low[a] := Ya;
    up[a]  := Ya;

{ --- set t0 to a pre-determined critical value ------------------- }

    lambda_hat := Ya/a;

    max_t0 := cp;

{ --- get C(1)  ----------------------------------------------------- }

    for Y1 := low[1] to up[1] do begin
      C1[Y1].p        := exp( -LFact[Y1] );
      C1[Y1].feasible := True;

      t := (lambda_hat - Y1/1) / sqrt((1/1- 1/a)*lambda_hat);

      if (t<max_t0) then Fp1[Y1] := 1
                    else Fp1[Y1] := 0;
    end;

{ --- get C(k+1) and then F(Y[k+1],t0) (k=1,..,(a-1)) --------------- }

    for k:=1 to (a-1) do begin              { === start of loop k === }

  { ---  Step 1 : get ranges of Y[k] and Y[k+1]  --- }

      low_Y1 := low[k];    up_Y1  := up[k];
      low_Y2 := low[k+1];  up_Y2  := up[k+1];

  { ---  Step 2 : calculate C(k+1)  --- }

      for Y2 := low_Y2 to up_Y2 do begin
        sum := 0;

        for Y1 := low_Y1 to up_Y1 do begin
          if feasible(Y1, Y2) and C1[Y1].feasible then begin
            C2[Y2].feasible := True;
            sum := sum + exp( Ln(C1[Y1].p) - LFact[Y2-Y1] );
          end;
        end;

        C2[Y2].p := sum;
      end;

  { ---  Step 3 : calculate F(Y[k+1],t0)  --- }

      for Y2 := low_Y2 to up_Y2 do begin
        if ((k+1)=a) then
          t := -infinity
        else 
          t := (lambda_hat - Y2/(k+1)) / sqrt((1/(k+1)- 1/a)*lambda_hat);

        if (t<max_t0) then begin
          sum := 0;

          for Y1 := low_Y1 to up_Y1 do
            if feasible(Y1, Y2) then 
              sum := sum + Fp1[Y1] * fk( Y1, Y2 );

          Fp2[Y2] := sum;
        end else
          Fp2[Y2] := 0;
      end;

  { --- Step 4 : copy C(k+1) & Fp(k+1) for the next iteration --- }

      if ((k+1) < a) then
        for J:=0 to max_factorial do begin
          C1[J].p:=C2[J].p; C1[J].feasible:=C2[J].feasible; Fp1[J]:=Fp2[J];
          C2[J].p:=0;       C2[J].feasible:=False;          Fp2[J]:=0;
        end;
    end;                                      { === end of loop k === }

    get_F := Fp2[Ya];
  end;                                         { --- end of get_F --- }

{ ======  M A I N    P A R T  ======================================= }

begin
  for J:=0 to max_sequence do y[J] := 0;

{ --- calculate log factorials -------------------------------------- }

                               Lfact[0] := 0;
  for J:=1 to max_factorial do Lfact[J] := Lfact[J-1] + Ln( J );

{ ---  read data   -------------------------------------------------- }

  write('Enter data file name ==>> '); readln( filename );

  if FileExists( filename ) then
    begin
      assign( fi, filename);                           reset(fi);
      assign( fo, filename + '_critical_point.csv' );  rewrite(fo);
    end
  else
    begin
      writeln;
      writeln('Error : file NOT found. Please try it again!');
      goto last_line;
    end;

  J  := 0;
  Ya := 0;

  repeat
    J := J+1;
    readln ( fi, y[J] );
    Ya := Ya + y[J];
  until eof(fi) or (J >= max_sequence);

  a := J;  { --- a is the length of the observed sequence y[] --- }

  if not eof(fi) then begin
    writeln('Warning: the length of the sequence is greater than max.' );
    readln;
    goto last_line;
  end;

  if (Ya > max_factorial) then begin
    writeln('Warning: the sum of yk is greater than max(1754).' );
    readln;
    goto last_line;
  end;

  close(fi);

  write('Enter alpha (0.05, 0.1, 0.2, ...) ==>> ');   readln( alpha );

  writeln( fo, 'a,',     a:4 );
  writeln( fo, 'Ya,',    Ya:4 );
  writeln( fo, 'alpha,', alpha:10:4 );
  writeln( fo );

  writeln;
  writeln( 'Now calculating ........ ' );
  writeln;

{ ---  find critical point by the bisection method ---------------------- }

  cp1 := 1.0;
  cp2 := 4.0;

  repeat
    alpha1 := 1 - get_F( a, cp1, y );
    alpha2 := 1 - get_F( a, cp2, y );

    if (alpha1<alpha) then cp1 := 0.5 * cp1;
    if (alpha2>alpha) then cp2 := 2.0 * cp2;
  until (alpha1>alpha) and (alpha2<alpha);

  writeln('   J       cp1       cp2        cp    alpha1    alpha2     alpha');
  writeln( fo, 'J, cp1, cp2, cp, alpha1, alpha2, alpha');

  J := 0;

  repeat
    J  := J+1;
    cp := (cp1+cp2)/2.0;
    p  := 1 - get_F( a, cp, y );

    if (p-alpha)>0 then cp1 := cp
                   else cp2 := cp;

    alpha1 := 1 - get_F( a, cp1, y );
    alpha2 := 1 - get_F( a, cp2, y );

    writeln(     J:4,
                 cp1:10:6,    cp2:10:6,    cp:10:6, 
                 alpha1:10:6, alpha2:10:6, p:10:6 );
    writeln( fo, J:4, ',',
                 cp1:10:6,    ',', cp2:10:6,    ',', cp :10:6, ',',
                 alpha1:10:6, ',', alpha2:10:6, ',', p:10:6 );

  until (p<alpha) and ((abs(p-alpha)<epsilon2) or (J>=30));

  cp := cp + epsilon;

  writeln;
  writeln( 'critical point = ', cp:8:5 );
  writeln( fo );
  writeln( fo, 'critical point,', cp:8:5 );

{ ---  ending  ------------------------------------------------------ }

  close(fo);

last_line: 
  writeln;
  write('Hit ENTER key !');
  readln;
end.
