program Max_acc_t1_Poisson(input, output);

{$APPTYPE CONSOLE}

uses SysUtils, Math;

{ ------------------------------------------------------------------- }
{   max acc t1 for Poisson sequence (step change-point model)         }
{     1.p-value  2.power  3.confidence set    @@@@@@            }
{ ------------------------------------------------------------------- }
{   This program can be compiled by Delphi(R).                        }
{   The latest version and the executable program are available at    }
{      https://corec.meisei-u.ac.jp/labs/hirotsu/                     }
{ ------------------------------------------------------------------- }
{   Critical points for power calculation are available at the site,  }
{   or use Ver.2 which searches the critical point automatically.     }
{ ------------------------------------------------------------------- }
{                  Notes and Limitations                              }
{ ------------------------------------------------------------------- }
{   Global variables C1 & C2 are used to avoid stack overflow.        }
{   Maximum length of Poisson sequence is 100.                        }
{   Extended precision is used                                        }
{   & the maximum number of total frequency is 1754.                  }
{      (If you use double precision, max number is 170.)              }
{ ------------------------------------------------------------------- }
{                                 written by Hal.Tsuruta  2017/04/24  }
{ ------------------------------------------------------------------- }

{ --- Declarations -------------------------------------------------- }

  const  max_sequence  =                 100;
         max_factorial =                1754;      { max of factorial }
         epsilon       =           0.0000001;
         infinity      =  1.797693134862E308;

  type array_type        = array[0..max_sequence]  of integer;
       array_type_R      = array[0..max_sequence]  of extended;
       array_type_long_R = array[0..max_factorial] of extended;

       Ck_type = record
                   p        : extended;
                   feasible : boolean;
                 end;

       option_type = ( get_p_value, get_power, get_CI );

  var  fi,   fo  : text;
       filename  : string;
       option    : option_type;

       y         : array_type;
       C1,  C2   : array[0..max_factorial] of Ck_type; { global var }
       LFact     : array_type_long_R;

       J,
       K0,  n_option,
       a,   Ya,
       change_point   : integer;

       p,   delta,
       critical_point : extended;

  label last_line;

{ ------------------------------------------------------------------- }
     function Imin2( x1, x2 : integer ) : integer;
{ ------------------------------------------------------------------- }
     begin
       if x1 < x2 then Imin2 := x1
                  else Imin2 := x2;
     end;

{ ------------------------------------------------------------------- }
     function feasible( Y1, Y2 : integer ) : boolean;
{ ------------------------------------------------------------------- }
     var yk : integer;

     begin
       yk := Y2 - Y1;

       if (yk >= 0) then feasible := True
                    else feasible := False;
     end;

{ ------------------------------------------------------------------- }
     function fk( k,  Y1,  Y2,
                  change_point : integer;
                  delta        : extended ) : extended;
{ ------------------------------------------------------------------- }
{         NB: C1 & C2 are gloabal variables                           }
{ ------------------------------------------------------------------- }
     var yk : integer;
         p  : extended;

     begin
       yk := Y2 - Y1;

       if feasible(Y1, Y2) and ((C1[Y1].feasible and C2[Y2].feasible)) then
         p := exp( Ln(C1[Y1].p) - Ln(C2[Y2].p) - LFact[yk] )
       else
         p := 0;

       if (k=change_point) then fk := p * exp(-delta*Y1)
                           else fk := p;
     end;

{ ------------------------------------------------------------------- }
     function get_F( a, K0,
                     change_point   : integer;
                     critical_point,
                     delta          : extended;
                     y              : array_type ) : extended;
{ ------------------------------------------------------------------- }
  var k,   J,
      Y1,  Y2,   Ya,
      low_Y1, up_Y1,
      low_Y2, up_Y2,
      max_k0          : integer;

      max_t0,     t,
      lambda_hat, sum : extended;

      Yk,  low,  up   : array_type;
      E0,  V0,   t0   : array_type_R;
      Fp1, Fp2        : array_type_long_R;

{ ----  Start of get_F  --------------------------------------------- }

  begin
    for J:=0 to max_factorial do begin
      C1[J].p:=0;  C1[J].feasible:=False;  Fp1[J]:=0;
      C2[J].p:=0;  C2[J].feasible:=False;  Fp2[J]:=0;
    end;

{ --- get accumulated sum Yk's of the observed sequence y ----------- }

                     Yk[0] := 0;
    for k:=1 to a do Yk[k] := Yk[k-1] + y[k];
                     Ya    := Yk[a];

{ --- set the range of Yk's ----------------------------------------- }

    for k:=1 to (a-1) do begin
      low[k] :=  0;
      up[k]  := Ya;
    end;

    low[a] := Ya;
    up[a]  := Ya;

    if (K0>0) then begin
      low[K0] := Yk[K0];
      up[K0]  := Yk[K0];
    end;

{ --- calculate t0 based on the observed sequence ------------------- }

    if (change_point=0) and (K0<=1) then begin
      writeln( fo );
      writeln( fo, 'k,y[k],Yk[k],E[k],V[k],t[k]' );

      writeln;
      writeln( '     k     y[k]    Yk[k]     E[k]        V[k]        t[k]' );
      writeln( '------------------------------------------------------------' );
    end;

    max_t0     := -infinity;
    max_k0     := 0;
    lambda_hat := Ya/a;

    if (change_point>0) then max_t0 := critical_point
    else begin
      for k:=1 to (a-1) do begin
        E0[k] := lambda_hat - Yk[k]/k;
        V0[k] := (1/k-1/a)*(lambda_hat);
        t0[k] := E0[k] / sqrt(V0[k]);

        if (t0[k] > max_t0) then begin
          max_t0 := t0[k];
          max_k0 := k;
        end;

        if (change_point=0) and (K0<=1) then begin
          writeln( fo,  k:8, ',',
                        y[k]:4,    ',',
                        Yk[k]:6,   ',',
                        E0[k]:10:4,',',
                        V0[k]:10:4,',',
                        t0[k]:10:4      );
          writeln(      k:6,       '  ',
                        y[k]:6,    '  ',
                        Yk[k]:6,   '  ',
                        E0[k]:10:4,'  ',
                        V0[k]:10:4,'  ',
                        t0[k]:10:4      );
        end;
      end;
    end;

    max_t0 := max_t0 - epsilon;

    if (change_point=0) and (K0<=1) then begin
      writeln( fo );
      writeln( fo,  'max_t0 =,', max_t0:10:4, ',  at K =,', max_k0:6    );
      writeln( fo );

      writeln( '------------------------------------------------------------' );
      writeln;
      writeln( 'max_t0 =', max_t0:10:4, ' at K = ', max_k0:3 );
      writeln;
    end;

{ --- get C(1)  ----------------------------------------------------- }

    for Y1 := low[1] to up[1] do begin
      C1[Y1].p        := exp( -LFact[Y1] );
      C1[Y1].feasible := True;

      if (K0=1) then t := -infinity
                else t := (lambda_hat - Y1/1) / sqrt((1/1- 1/a)*lambda_hat);

      if (t<max_t0) then Fp1[Y1] := 1
                    else Fp1[Y1] := 0;
    end;

{ --- get C(k+1) and then F(Y[k+1],t0) (k=1,..,(a-1)) --------------- }

    for k:=1 to (a-1) do begin              { --- start of loop k --- }

  { ---  Step 1 : get ranges of Y[k] and Y[k+1]  --- }

      low_Y1 := low[k];    up_Y1  := up[k];
      low_Y2 := low[k+1];  up_Y2  := up[k+1];

  { ---  Step 2 : calculate C(k+1)  --- }

      for Y2 := low_Y2 to up_Y2 do begin
        sum := 0;

        for Y1 := low_Y1 to up_Y1 do begin
          if feasible(Y1, Y2) and C1[Y1].feasible then begin
            C2[Y2].feasible := True;

            if (k=change_point) then
              sum := sum + exp( Ln(C1[Y1].p) - LFact[Y2-Y1]) * exp(-delta*Y1)
            else
              sum := sum + exp( Ln(C1[Y1].p) - LFact[Y2-Y1] );
          end;
        end;

        C2[Y2].p := sum;
      end;

  { ---  Step 3 : calculate F(Y[k+1],t0)  --- }

      for Y2 := low_Y2 to up_Y2 do begin
        if ((k+1)=a) or ((k+1)=K0) then
          t := -infinity
        else 
          t := (lambda_hat - Y2/(k+1)) / sqrt((1/(k+1)- 1/a)*lambda_hat);

        if (t<max_t0) then begin
          sum := 0;

          for Y1 := low_Y1 to up_Y1 do
            if feasible(Y1, Y2) then
              sum := sum + Fp1[Y1] * fk( k, Y1, Y2, change_point, delta );

          Fp2[Y2] := sum;
        end else
          Fp2[Y2] := 0;
      end;

  { --- Step 4 : copy C(k+1) & Fp(k+1) for the next iteration --- }

      if ((k+1) < a) then
        for J:=0 to max_factorial do begin
          C1[J].p:=C2[J].p; C1[J].feasible:=C2[J].feasible; Fp1[J]:=Fp2[J];
          C2[J].p:=0;       C2[J].feasible:=False;          Fp2[J]:=0;
        end;
    end;                                      { --- end of loop k --- }

    get_F := Fp2[Ya];
  end;                                         { --- end of get_F --- }

{ ------  M A I N    P A R T  --------------------------------------- }

begin
  K0             := 0;
  delta          := 0;
  change_point   := 0;
  critical_point := -infinity;

  for J:=0 to max_sequence do y[J] := 0;

{ --- calculate log factorials -------------------------------------- }

                               Lfact[0] := 0;
  for J:=1 to max_factorial do Lfact[J] := Lfact[J-1] + Ln( J );

{ ---  read data   -------------------------------------------------- }

  write('Enter option (1.p-value 2.power 3.confidence set) ==>> ');
  readln( n_option );

       if n_option=2 then option := get_power
  else if n_option=3 then option := get_CI
  else                    option := get_p_value;

  write('Enter data file name ==>> '); readln( filename );

  if FileExists( filename ) then
    begin
      assign( fi, filename);          reset(fi);
      assign( fo, filename+'.csv' );  rewrite(fo);
    end
  else
    begin
      writeln;
      writeln('Error : file NOT found. Please try it again!');
      goto last_line;
    end;

  J  := 0;
  Ya := 0;

  repeat
    J := J + 1;
    readln ( fi, y[J] );
    Ya := Ya + y[J];
  until eof(fi) or (J >= max_sequence);

  a := J;  { --- a is the length of the observed sequence y[] --- }

  if not eof(fi) then begin
    writeln('Warning: the length of the sequence is greater than max.' );
    goto last_line;
  end;

  if (Ya > max_factorial) then begin
    writeln('Warning: the sum of yk is greater than max(1754).' );
    goto last_line;
  end;

  close(fi);

  if (option=get_power) then begin
    write('Enter change point ==>> ');     readln( change_point );
    write('Enter critical point ==>> ');   readln( critical_point );
  end;

  writeln;
  writeln( 'Now calculating ........ ' );
  writeln;

{ ---  top level   -------------------------------------------------- }

  case option of
    get_p_value :
      begin
        p := 1 - get_F( a, 0, 0, 0, 0, y );

        writeln(     'p-value: ', p:10:6 );
        writeln( fo, 'p-value:,', p:10:6 );
      end;
 
    get_power :
      begin
        writeln( '   delta        p' );
        writeln( '  ---------------------' );
        writeln( fo, 'delta,p' );

        for J:=-200 to 200 do begin
          delta := J / 100;
          p := 1 - get_F( a, 0, change_point, critical_point, delta, y );

          writeln(     delta:8:2, '  ', p:10:6 );
          writeln( fo, delta:8:2, ',',  p:10:6 );
        end;

        writeln( '  ---------------------' );
      end;

    get_CI :
      begin
        for K0:=1 to (a-1) do begin
          p := 1 - get_F( a, K0, 0, 0, 0, y );

          if (K0=1) then begin
            writeln( '      K0        p' );
            writeln( '    -----------------------' );
            writeln( fo, 'K0,p' );
          end;

          write( K0:8, '  ', p:10:6 );
               if (p>=0.1)  then writeln( '   **' )
          else if (p>=0.05) then writeln( '   *' )
          else                   writeln;

          write( fo, K0:8, ',',  p:10:6 );
               if (p>=0.1)  then writeln( fo, ',**' )
          else if (p>=0.05) then writeln( fo, ',*' )
          else                   writeln( fo );
        end;

        writeln( '    -----------------------' );
      end;
  end;

{ ---  ending  ------------------------------------------------------ }

  close(fo);

last_line: 
  writeln;
  write('Hit ENTER key !');
  readln;
end.
